/*	freeems.h

	Copyright 2008 Fred Cooke

	This file is part of the FreeEMS project.

	FreeEMS software is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	FreeEMS software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with any FreeEMS software.  If not, see <http://www.gnu.org/licenses/>.

	We ask that if you make any changes to this file you send them upstream to us at admin@diyefi.org

	Thank you for choosing FreeEMS to run your engine! */

/* Header file multiple inclusion protection courtesy eclipse Header Template	*/
/* and http://gcc.gnu.org/onlinedocs/gcc-3.1.1/cpp/ C pre processor manual		*/
#ifndef FILE_FREEEMS_H_SEEN
#define FILE_FREEEMS_H_SEEN

#include "memory.h"
#include "9s12xdp512.h"

/* Common macro definitions across the project */
/* Boolean */
#define TRUE 1
#define FALSE 0

/* 8 bit values */
#define ZEROS 0x00
#define ONES 0xFF

/* 16 bit values */
#define ZEROS16 0x0000
#define ONES16 0xFFFF

#define LONGHALF 0x80000000

/* Individual bits WARNING, do not not these for notted versions, use the notted ones below instead : http://supp.iar.com/Support/?note=12582&from=search+result */
#define BIT0		0x01	/* 1st bit	= 1		*/
#define BIT1		0x02	/* 2nd bit	= 2		*/
#define BIT2		0x04	/* 3rd bit	= 4		*/
#define BIT3		0x08	/* 4th bit	= 8		*/
#define BIT4		0x10	/* 5th bit	= 16	*/
#define BIT5		0x20	/* 6th bit	= 32	*/
#define BIT6		0x40	/* 7th bit	= 64	*/
#define BIT7		0x80	/* 8th bit	= 128	*/

#define BIT0_16		0x0001	/* 1st bit	= 1		*/
#define BIT1_16		0x0002	/* 2nd bit	= 2		*/
#define BIT2_16		0x0004	/* 3rd bit	= 4		*/
#define BIT3_16		0x0008	/* 4th bit	= 8		*/
#define BIT4_16		0x0010	/* 5th bit	= 16	*/
#define BIT5_16		0x0020	/* 6th bit	= 32	*/
#define BIT6_16		0x0040	/* 7th bit	= 64	*/
#define BIT7_16		0x0080	/* 8th bit	= 128	*/

#define BIT8_16		0x0100	/* 9th bit	= 256	*/
#define BIT9_16		0x0200	/* 10th bit	= 512	*/
#define BIT10_16	0x0400	/* 11th bit	= 1024	*/
#define BIT11_16	0x0800	/* 12th bit	= 2048	*/
#define BIT12_16	0x1000	/* 13th bit	= 4096	*/
#define BIT13_16	0x2000	/* 14th bit	= 8192	*/
#define BIT14_16	0x4000	/* 15th bit	= 16384	*/
#define BIT15_16	0x8000	/* 16th bit	= 32768	*/

/* Individual bits NOTTED : http://supp.iar.com/Support/?note=12582&from=search+result */
#define NBIT0		0xFE	/* 1st bit	= 1		*/
#define NBIT1		0xFD	/* 2nd bit	= 2		*/
#define NBIT2		0xFB	/* 3rd bit	= 4		*/
#define NBIT3		0xF7	/* 4th bit	= 8		*/
#define NBIT4		0xEF	/* 5th bit	= 16	*/
#define NBIT5		0xDF	/* 6th bit	= 32	*/
#define NBIT6		0xBF	/* 7th bit	= 64	*/
#define NBIT7		0x7F	/* 8th bit	= 128	*/

#define NBIT0_16	0xFFFE	/* 1st bit	= 1		*/
#define NBIT1_16	0xFFFD	/* 2nd bit	= 2		*/
#define NBIT2_16	0xFFFB	/* 3rd bit	= 4		*/
#define NBIT3_16	0xFFF7	/* 4th bit	= 8		*/
#define NBIT4_16	0xFFEF	/* 5th bit	= 16	*/
#define NBIT5_16	0xFFDF	/* 6th bit	= 32	*/
#define NBIT6_16	0xFFBF	/* 7th bit	= 64	*/
#define NBIT7_16	0xFF7F	/* 8th bit	= 128	*/

#define NBIT8_16	0xFEFF	/* 9th bit	= 256	*/
#define NBIT9_16	0xFDFF	/* 10th bit	= 512	*/
#define NBIT10_16	0xFBFF	/* 11th bit	= 1024	*/
#define NBIT11_16	0xF7FF	/* 12th bit	= 2048	*/
#define NBIT12_16	0xEFFF	/* 13th bit	= 4096	*/
#define NBIT13_16	0xDFFF	/* 14th bit	= 8192	*/
#define NBIT14_16	0xBFFF	/* 15th bit	= 16384	*/
#define NBIT15_16	0x7FFF	/* 16th bit	= 32768	*/

/* TODO perhaps configure various ports and pins etc to have names such that the
 * code can stay untouched when being ported and just the header changed to suit?
 * definitely a good idea, but we'll see what karsten says before putting time
 * into it. something like : #define RPM_IN_MAIN IC0 etc etc
 */

/* Definitions for functions used across the project */
/* http://www.seattlerobotics.org/encoder/200109/encoder.htm section Assembly */
#define INLINE_ASSEMBLY(code) __asm__ __volatile__ (code)

/* GLOBAL Variables */
// TODO change any of these that need it to volatile!!!

/* Types summary
 * 
 * BEWARE : Be explicit!!
 * 					
 * char			8 bit (defaults to unsigned, but always specify signed/unsigned anyway)
 * short		16 bit (defaults to signed, but always specify signed/unsigned anyway)
 * int			16 bit DO NOT USE! (current compile flags make this 16 bits, but a change of flags could will change your program if you use this because they will all be 32 bit all of a sudden)
 * long			32 bit (defaults to signed, but always specify signed/unsigned anyway)
 * long long	64 bit (inefficient, avoid these, if using : defaults to signed, but always specify signed/unsigned anyway)
 * float		32 bit IEEE floating point numbers (inefficient, avoid these, used fixed point math)
 * double		64 bit IEEE floating point numbers (inefficient, avoid these, used fixed point math)
 */


/* Constant definitions : see global_constants.c for values */
/* These need to be changed if the timer period is changed at all */
/* The number of timer units it takes for the switch on scheduling code to run */
const unsigned short injectorSwitchOnCodeTime;
/* The number of timer units it takes for the switch off scheduling code to run */
const unsigned short injectorSwitchOffCodeTime;
/* The maximum a requested pulsewidth can be before it is truncated to this amount */
const unsigned short injectorMaximumPulseWidth;
/* The minimum a requested pulsewidth can be before it is not switched on at all */
const unsigned short injectorMinimumPulseWidth;


/* Ignition limits */
/* Ignition maximum dwell in timer units */
const unsigned short ignitionMaximumDwell;

/* Ignition minimum dwell in timer units */
const unsigned short ignitionMinimumDwell;

/* Ignition maximum delay post schedule tooth in timer units */
const unsigned short ignitionMaximumDelayToDwellStartAfterTooth;


/*  */
const unsigned short leadingEdgePrimaryRPMInputCodeTime; 
/*  */
const unsigned short trailingEdgePrimaryRPMInputCodeTime; 
/*  */
const unsigned short leadingEdgeSecondaryRPMInputCodeTime; 
/*  */
const unsigned short trailingEdgeSecondaryRPMInputCodeTime; 

/*  */
const unsigned short baudDivisor115p2;

/* Counters (all require init to zero) */
unsigned char UISRCounter;							/* Counter to ensure we aren't accidentally triggering unused ISRs */
unsigned char lowVoltISRCounter;					/* Counter for low voltage conditions */
unsigned char readingCounter;						/* Counter for timeout of readings */
unsigned char lostCrankSyncCounter;					/* Counter for number of lost crank syncs */
unsigned char lostCamSyncCounter;					/* Counter for number of lost cam syncs */
unsigned char lostRPMValidityCounter;				/* Counter for number of lost RPM validity events */
/* Free running counters for number of teeth seen such that tooth timing can be used to reconstruct the signal at lower rpm */
unsigned short primaryTeethCounter;	
unsigned short secondaryTeethCounter;

/* ECT IC extension variable (init not required, don't care where it is, only differences between figures) */
unsigned short timerExtension;						/* Increment for each overflow of the main timer, allows finer resolution and longer time period */
/* section 10.3.5 page 290 68hc11 reference manual e.g. groups.csail.mit.edu/drl/courses/cs54-2001s/pdf/M68HC11RM.pdf */


/* Real Time Clocks (all require init to zero) */
unsigned short realTimeCounterMain;					/* Variable to count RTI executions, 0.125ms exactly	*/
unsigned short realTimeCounterMillis;				/* Variable to count milliseconds exactly				*/
unsigned short realTimeCounterTenths;				/* Variable to count tenths of a second exactly			*/
unsigned short realTimeCounterSeconds;				/* Variable to count seconds exactly					*/
unsigned short realTimeCounterMinutes;				/* Variable to count minutes exactly					*/
unsigned char millisToTenths;						/* Rollover variable for counting tenths				*/
unsigned char tenthsToSeconds;						/* Rollover variable for counting seconds				*/
unsigned char secondsToMinutes;						/* Rollover variable for counting minutes				*/


/* Setting variables (must be inited with useful values) */
unsigned char ports;								/* How many groups of injectors we are firing			*/
unsigned char coils;								/* How many coils we are firing							*/
unsigned char combustionEventsPerEngineCycle;		/* How many power producing events per engine cycle		*/
unsigned char revolutionsPerEngineCycle;			/* Rotary = 1, 2 Stroke = 1, 4 Stroke = 2				*/
unsigned short capacityOfAirPerCombustionEvent;		/* How much air it swallows per power producing event	*/
unsigned short perPrimaryInjectorChannelFlowRate;	/* How much fuel flows per minute per channel			*/
unsigned short perSecondaryInjectorChannelFlowRate;	/* How much fuel flows per minute per channel			*/
unsigned char primaryTeeth;							/* How many teeth are on the crank signal including the missing ones if any (eg. 36-1 primary = 36 not 35) */
unsigned char missingTeeth;							/* Number sequentially removed from primary teeth (eg. 36-1 missing = 1) */ 
unsigned char readingTimeout;						/* How often an ADC reading MUST occur */


/* State variables : 0 = false (don't forget to change the init mask to suit!) */
unsigned short coreStatusA;		/* Each bit represents the state of some core parameter, masks below */
#define RPM_VALID BIT1_16		/*  1 Whether we are sure rpm is what the variable says (used to inject fuel without ignition below the threshold rpm) */
#define PRIMARY_SYNC BIT2_16	/*  2 Wasted spark/Semi sequential */
#define SECONDARY_SYNC BIT3_16	/*  3 COP/Full sequential */
#define ENGINE_PHASE BIT4_16	/*  4 For COP/Sequential, which revolution we are in, first or second */
#define FUEL_CUT BIT5_16		/*  5 Remove injection completely */
#define HARD_SPARK_CUT BIT6_16	/*  6 Remove ignition completely */
#define SOFT_SPARK_CUT BIT7_16	/*  7 Remove ignition events round robin style */
#define SPARK_RETARD BIT8_16	/*  8 Retard ignition in RPM dependent way */
#define STAGED_REQUIRED BIT9_16	/*  9 Whether we are firing the staged injectors */
#define CALC_FUEL_IGN BIT10_16	/* 10 Whether fuel and ignition require calculation (i.e. variables have been updated) */
#define FORCE_READING BIT11_16	/* 11 Flag to force ADC sampling at low rpm/stall */
#define CALCS_VALID BIT12_16	/* 12 Whether the calculated outputs are valid or not */
#define COREA13 BIT13_16		/* 13 */
#define COREA14 BIT14_16		/* 14 */
#define COREA15 BIT15_16		/* 15 */
#define COREA16 BIT16_16		/* 16 */

/* Which bank to read from or write to at any time */
unsigned char inputBank;
unsigned char outputBank;
unsigned char outputBankInjectionOffset;
unsigned char outputBankIgnitionOffset;

/* Settings variables : 0 = false */ 
unsigned short coreSettingsA;	/* Each bit represents the state of some core setting, masks below and above where the same one is used */
//#define COREA1 BIT1_16			/*  1 */
#define PRIMARY_POLARITY BIT2_16	/*  2 1 = high teeth 0 = low teeth */
#define SECONDARY_POLARITY BIT3_16	/*  3 1 = high teeth 0 = low teeth */
//#define COREA4 BIT4_16			/*  4 */
//#define FUEL_CUT BIT5_16			/*  5 Remove injection completely */
//#define HARD_SPARK_CUT BIT6_16	/*  6 Remove ignition completely */
//#define SOFT_SPARK_CUT BIT7_16	/*  7 Remove ignition events round robin style */
//#define SPARK_RETARD BIT8_16		/*  8 Retard ignition in RPM dependent way */
#define STAGED_ON BIT9_16			/*  9 Whether we are firing the staged injectors */
#define STAGED_START BIT10_16		/* 10 1 = Fixed start 0 = Scheduled start */
#define STAGED_END BIT11_16			/* 11 1 = Fixed end 0 = Scheduled end */
//#define COREA12 BIT12_16			/* 12 */
//#define COREA13 BIT13_16			/* 13 */
//#define COREA14 BIT14_16			/* 14 */
//#define COREA15 BIT15_16			/* 15 */
//#define COREA16 BIT16_16			/* 16 */


/* Flag registers, init to zero required */
unsigned char mainOn;				/* Keep track of where we are at for possible use as multi interrupt per injection */
unsigned short dwellOn;				/* Keep track of ignition output state */
unsigned char stagedOn;				/* Ensure we turn an injector off again if we turn it on. */
unsigned char selfSetTimer;			/* Set the start time of injection at the end of the last one in the channels ISR instead of the input ISR */
unsigned char rescheduleFuelFlags;	/* Pulse width is probably longer than engine cycle so schedule a restart at the next start time */


/* Engine runtime properties (inits???) TODO */
unsigned short primaryPulsesPerSecondaryPulse; /* Type short because of nissan style cam wheels (char would do for other types) */
unsigned short primaryPulsesPerSecondaryPulseBuffer; /* Type short because of nissan style cam wheels (char would do for other types) */
unsigned short primaryLeadingEdgeTimeStamp; /* Store the timestamp of the leading edge during a pulse */
unsigned short primaryTrailingEdgeTimeStamp; /* Store the timestamp of the leading edge during a pulse */
unsigned short timeBetweenSuccessivePrimaryPulses; /* This number equates to the speed of the engine */
unsigned short timeBetweenSuccessivePrimaryPulsesBuffer; /* This number equates to the speed of the engine */
unsigned short lastPrimaryPulseTimeStamp; /* Store the timer value of the each pulse here before exiting the ISR */
unsigned long engineCyclePeriod;
unsigned long lastSecondaryOddTimeStamp;


/* Output variables (init not required) TODO tidy this up significantly */
unsigned short totalPulseWidth;
unsigned short totalDwell;
unsigned short totalAngleAfterReferenceInjection;
unsigned short totalAngleAfterReferenceIgnition;


// test vars
unsigned char portHDebounce; // init to zero required

#define IN_OUT_BANKS 2
#define INJECTION_CHANNELS 6							/* How many injection channels the code should support */
#define IGNITION_CHANNELS 12							/* How many ignition channels the code should support */

/* TODO tidy this up significantly */
unsigned short rpm[IN_OUT_BANKS];			/* Revolutions Per Minute			: 0		- 65,536		(1 RPM)					*/
unsigned short clt[IN_OUT_BANKS];			/* Coolant Temperature				: 0.0	-    511.5		(0.5 Celsius (/2 -100))		*/
unsigned short map[IN_OUT_BANKS];			/* Manifold Absolute Pressure		: 0.0	-    511.5		(0.5 kPa (/2))			*/
unsigned short iat[IN_OUT_BANKS];			/* Inlet Air Temperature			: 0.0	-    102.3		(0.1 Celsius (/10,-50))	*/
unsigned short tps[IN_OUT_BANKS];			/* Throttle Position Sensor			: 0.0	-    102.3		(0.1 % (/10))			*/
unsigned short ego[IN_OUT_BANKS];			/* Exhaust Gas Oxygen				: 0.000	-      2.046	(0.002 lambda (/500))	*/
unsigned short spareADC5[IN_OUT_BANKS];		/* Spare ADC port 5	*/
unsigned short spareADC6[IN_OUT_BANKS];		/* Spare ADC port 6	*/
unsigned short spareADC7[IN_OUT_BANKS];		/* Spare ADC port 7	*/

/* TODO add in the other 8 ADC channels */
//

/* Ignition stuff */

#ifndef NO_CONST_ARRAYS

/*extern*/ const unsigned short dwellStartMasks[IGNITION_CHANNELS];
/*extern*/ const unsigned short ignitionMasks[IGNITION_CHANNELS];
/*extern*/ const unsigned char injectorMainOnMasks[INJECTION_CHANNELS];
/*extern*/ const unsigned char injectorMainOffMasks[INJECTION_CHANNELS];
/*extern*/ const unsigned char injectorMainEnableMasks[INJECTION_CHANNELS];
/*extern*/ const unsigned char injectorMainDisableMasks[INJECTION_CHANNELS];
/*extern*/ const unsigned char injectorMainGoHighMasks[INJECTION_CHANNELS];
/*extern*/ const unsigned char injectorMainGoLowMasks[INJECTION_CHANNELS];

#endif

/*extern*/ volatile unsigned short * volatile injectorMainTimeRegisters[INJECTION_CHANNELS];
/*extern*/ volatile unsigned char * volatile injectorMainControlRegisters[INJECTION_CHANNELS];  

// ignition experimentation stuff
unsigned char dwellQueueLength;				/* 0 = no dwell pending start, 1 = single event scheduled, 2 = one scheduled, and one in the queue, etc */
unsigned char ignitionQueueLength;			/* 0 = no spark event pending, 1 = single event scheduled, 2 = one scheduled, and one in the queue, etc */
unsigned short currentDwell[IN_OUT_BANKS];	/* Current dwell to be used. */
unsigned char nextDwellChannel;				/* Which one to bang off next */
unsigned char nextIgnitionChannel;			/* Which one to bang off next */
unsigned short ignitionAdvances[IGNITION_CHANNELS * 2]; // Uses channel + offset to have two values at any time
unsigned short queuedDwellOffsets[IGNITION_CHANNELS]; // Uses next channel as index
unsigned short queuedIgnitionOffsets[IGNITION_CHANNELS]; // Uses next channel as index

#define DWELL_ENABLE BIT0
#define DWELL_DISABLE NBIT0
#define IGNITION_ENABLE BIT1
#define IGNITION_DISABLE NBIT1


/* Timer holding vars (init not required) */
unsigned short injectorMainStartTimesHolding[INJECTION_CHANNELS];
unsigned long injectorMainEndTimes[INJECTION_CHANNELS];

// TODO make these names consistent
/* Code time to run variables (init not required) */
unsigned short injectorCodeOpenRuntimes[INJECTION_CHANNELS];
unsigned short injectorCodeCloseRuntimes[INJECTION_CHANNELS];

/* individual channel pulsewidths (init not required) */
unsigned short injectorMainPulseWidths[INJECTION_CHANNELS * 2];
unsigned short injectorStagedPulseWidths[INJECTION_CHANNELS * 2];

/* Channel latencies (init not required) */
unsigned short injectorCodeLatencies[INJECTION_CHANNELS];

#else
	/* let us know if we are being untidy with headers */
	#warning "Header file FREEEMS_H seen before, sort it out!"
/* end of the wrapper ifdef from the very top */
#endif


/* TODO %%%%%%%%%%%%%%%%%%%%%%%%%%% use structs to hold appropriate things (settings etc) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
/* TODO for noobs like me : http://www.space.unibe.ch/comp_doc/c_manual/C/SYNTAX/struct.html
 * TODO                     http://en.wikipedia.org/wiki/Composite_type
 * TODO                     http://www.cs.usfca.edu/~wolber/SoftwareDev/C/CStructs.htm
 * TODO                     http://sandbox.mc.edu/~bennet/cs220/codeex/struct_c.html
 * TODO 
 * TODO
 * TODO put them in a separate C file with appropriate headers in here or its own header maybe with #include here
 * TODO
 * TODO typedef struct { structure of vars etc } box;
 * TODO
 * TODO box b;
 * TODO b.var = value;
 * TODO
 * TODO struct fred {
 * TODO short s;
 * TODO char str[26];
 * TODO float f;
 * TODO };
 * TODO 
 * TODO struct fred f1 = { 40, "Howdy", 5.498 };
 * TODO struct fred f2, *fp;
 * TODO 
 * TODO f1.f = 9834.0389843; // floating point                                              
 * TODO f1.s = 65536; // short max
 * TODO f1.str[25] = 'z'; // char character
 * TODO f1.str[24] = 255; // char number max
 * TODO f2.str = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
 * TODO           01234567890123456789012345
 * TODO
 * TODO fp->var
 * TODO equals
 * TODO (*fp).var
 * TODO 
 * TODO fp->f = 9834.0389843;                                              
 * TODO fp->s = 65536;                         
 * TODO fp->str = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
 * TODO           01234567890123456789012345
 * TODO                                              */
